! ===========================================================================================================
! integer function get_pqc.f90
! 
! -----------------------------------------------------------------------------------------------------------
!
! Revision history
! date     who    what
! 20170712 all    Release of qcplus v1.1 - many changes to all, some documented/tracked, some not
! 20160511 all    Initial release of qcplus v1.1.0
!
! ===========================================================================================================
! Notes
! This function selects the appropriate previous qc file, and then calls read_pqc to read
! the data needed for tests like speed check, grey list,salinity adjustment
!
! get_pqc finds previous qc files for the float being processed. It decrements the cycle,
! forms the name of the pqc file, and sees if that file exists (the "inquire" command).
! If it does not exist, it decrements the cycle again and re-tests. If it does exist, it
! calls "read_pqc" to get the information on 1) previous location for the speed check and
! 2) profile informatin for P,T,S for the jump check and frozen profile test.
!
! If the required information is NOT in the file (bad location and/or no profile), the process
! is reapeated, with one exception.  Only the location is extracted, profile informatin is
! NOT extracted for pqc files after the first.
!
! The recursive search continues for a period of DT_MAX_DAYS (a parameter set in data_arrays.f90).
! When this limit is exceeded, the launch position is used as teh location.
!
! ===========================================================================================================

integer function get_pqc(pathname)

use communication                       ! module
use data_arrays                         ! module

implicit none

! ===========================================================================================================
! varialbe/array declarations

character(len=*), intent(in) :: pathname ! Note that it is NOT USED (FIX this)

character(len=MAXPROFCHR) :: pqcfile
character(len=6)          :: pqcextn

character(len=256) :: prevqc, stem
character(len=3) :: typ
integer, parameter :: maxback = 3; ! maximum number of profiles to go back
integer :: ret = 0;
integer :: ios, ic,  nback;
integer :: MAXNBACK = 50;	! For most floats, about a year
logical :: file_exist
logical :: first_read
logical :: first_retry = .true.
character(len=10), parameter :: sub = 'get_pqc   '
integer :: read_pqc		! This function actually loads the data

! ===========================================================================================================
! executable statements

! Set the defaults now, in case no pqc file is found
FOUND_P_POS = .false.;
FOUND_P_PROF = .false.;
first_read = .true.;
found_pres_mrvsp = .false.; pres_mrvsp = FILL;
found_pqc_mrvsp = .false.; pqc_mrvsp = FILL;

get_pqc = FAILURE; 					! Assume failure

prev_good_lat = FILL; prev_good_lon = FILL; prev_good_posn_date = FILL14CH;

if(DBGIO) call dbg_enter(sub);
if(DEBUG)then; write(msg,'("at 1 pathname: ",a)')trim(pathname); call wrtdbg(sub,msg); endif
ic=len(trim(pathname))
stem=trim(pathname);

prevcyc=thiscyc; nback=0;
43 continue ! RETRY_CYCLE Retry point 1 - go back one cycle

if(prevcyc == 0) then
    write(msg,'("No pqc files were found, return FAILURE, launch position/date/time will be used for speed check")');
    call wrttrc(sub,msg); ! skip down to normal return
    go to 999;
endif
    
prevcyc=prevcyc - 1; nback = nback + 1;

! first, look for the same asc or dsc files as the inupt phy file
if(phy_descending)then; pqcextn = "D.qc"; typ = "dsc"; else; pqcextn = ".qc"; typ = "asc"; endif
pqcfile = compose_profname(sub,float_name,prevcyc,pqcextn);
prevqc = trim(pathname_aomlid_qcdir) // trim(pqcfile)
inquire(file=prevqc,exist=file_exist)
if(file_exist)then
    write(msg,'("The prevqc ",a3," file exists: ",a)')typ,trim(prevqc); call wrtlog(sub,msg);
else
    write(msg,'("The prevqc ",a3," file DOES NOT EXIST: ",a)')typ,trim(prevqc); call wrtlog(sub,msg);
    ! Look for the other file (always make this check)
    if(phy_descending)then; pqcextn = ".qc"; typ = "asc"; else; pqcextn = "D.qc"; typ = "dsc"; endif
    pqcfile = compose_profname(sub,float_name,prevcyc,pqcextn);
    prevqc = trim(pathname_aomlid_qcdir) // trim(pqcfile)
    inquire(file=prevqc,exist=file_exist)
    if(file_exist)then
	write(msg,'("The prevqc ",a3," file exists: ",a)')typ,trim(prevqc); call wrtlog(sub,msg);
    else
	write(msg,'("The prevqc ",a," file DOES NOT EXIST: ",a)')typ,trim(prevqc); call wrtlog(sub,msg);
	go to 43;	! RETRY_CYCLE
    endif
endif
if(DEBUG) write(DBG,'("get_pqc TEST We found a pqc file: ",a)')trim(prevqc)
if(DEBUG) write(DBG,'("DBG PQC  get_pqc  call read_pqc for file: ",a)')trim(prevqc);

pqc_mrvsp = FILL; found_pqc_mrvsp = .false.;

  ret = read_pqc(sub,prevqc,first_read);

if(DEBUG)then; write(msg,'("GET_PQC rtrn read_pqc, ret=",i2)')ret; call wrtdbg(sub,msg); endif

first_read = .false.;       ! read_pqc will NOT read profile unless it is the first read
if(ret == FAILURE)then
    if(DEBUG) write(DBG,'("DBG PQC  get_pqc ret=FAILURE")')
    if(DEBUG) write(DBG,'("DBG PQC  get_pqc  prev_posn_found/good: ",2l2)')prev_posn_found, prev_posn_good
    if(TIME_DIFF_SMALL)then	! TIME_DIFF_SMALL (set by DT_MAX_DAYS) means the time diff is < 365 days, continue search
	write(msg,'("previsous qcfile: ",a," POS NOT RETRIEVED")')trim(prevqc); call wrttrc(sub,msg);
	go to 43 ! Retry point
    else
	write(msg,'("previsous qcfile: ",a," POS NOT RETRIEVED, retries exhausted, return FAILURE")')trim(prevqc);
	call wrttrc(sub,msg);
    endif
else
    if(DEBUG) write(DBG,'("DBG PQC  get_pqc ret=SUCCESS")')
    if(DEBUG) write(DBG,'("DBG PQC  get_pqc  prev_posn_found/good: ",2l2)')prev_posn_found, prev_posn_good
    write(msg,'("previsous qcfile: ",a," RETRIEVED A POSITION")')trim(prevqc)
   
    if(.not. prev_posn_good)then
	write(msg,'("HOWEVER, the position was BAD, so continue the search")')
	if(DEBUG) write(DBG,'("DBG PQC  get_pqc  HOWEVER, the position was BAD, so continue the search")');
	go to 43 ! Retry point
    endif
endif

get_pqc = SUCCESS;				! Signal success

999 continue ! This is THE_END

if(DBGIO) call dbg_leave(sub,get_pqc);

return;
end function get_pqc
! ===========================================================================================================
