! ===========================================================================================================
! module ncbiop_arrays.f90
!
! -----------------------------------------------------------------------------------------------------------
!
! Revision history
! date     who    what
! 20180307 VH     Make functional a non-functional f90. Add all arrays and create/def/put subroutines to module
! 20170712 all    Release of qcplus v1.1 - many changes to all, some documented/tracked, some not
! 20160511 all    Initial release of qcplus v1.1.0
!
! ===========================================================================================================
! Notes:

! This module contains variable/array definitions used by write_ncbiop to create the nc Bio-ARGO profile file.
! ===========================================================================================================

module ncbiop_arrays

use communication                                                ! module
use data_arrays                                                  ! module
use netcdf_arrays                                                ! module
use ncprof_arrays                                                ! module

implicit none

! ===========================================================================================================
! moedule varialbe/array declarations
! ===========================================================================================================

integer :: N_VALUE; ! BioARGO dimension - how many values for a given measurement at each level (rang of frequencies, etc)
integer :: N_VALUE_dimid        ; integer :: N_VALUE_len                               ! 
integer, parameter :: N_VALUE1 = 1;      ! So far, we have no parameters with multiple values at each level (usually by frequency)


! Note that while DOXY is listed as the first parameter, DOXY is NOT read into the profbdata array, at this time.  
! Later, after v3.1 is tested enough, we might change this. especially if we also want to qc NITRACT.



! Used only in B0Argo ncbiop files
integer :: PARAMETER_DATA_MODE_id, PARAMETER_DATA_MODE_dims(NDIMS2); character(len=1) :: PARAMETER_DATA_MODE(MAXPRF,MAXBPAR);

integer :: BPRES_id, BDOXY_ids(NCARRS), BPAR_ids(5);
integer :: TEMP_DOXY_ids(5), FREQ_DOXY_ids(5), BPHA_DOXY_ids(5), TPHA_DOXY_ids(5), RPHA_DOXY_ids(5)

integer :: DOXY_profqc_id, BBPAR_profqc_ids(MAXBPAR);
real :: BDATA(MAXLEV); character(len=1) :: BFLAG(MAXLEV);
real :: FDATA(MAXLEV); character(len=1) :: FFLAG(MAXLEV); ! Only DOXY will get qc flags, use this for iprf>2
                       character(len=1) :: FNOTD(MAXLEV); ! QC flag cannot be determined
integer :: BSTATION_PARS_id      ,     BSTN_PARAMETERS_dims(ND3);        character(len=64) :: BSTN_PARAMETERS(MAXBPAR+2);
logical :: hasBDAT(MAXBPAR+2,MAXPRF);
character(len=64) :: BPARAMETER, BPARAMETERS(MAXBPAR+2,MAXPRF);
character(len=64) :: blank64 = "                                                                      ";
integer :: BPROF_QC_ids(MAXPAR);               ! nc ids for Profile_<>)qc

! ===========================================================================================================

contains

! ===========================================================================================================
! module functions/subroutines
! ===========================================================================================================
!

subroutine def_ncbiop_vars(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;

    character(len=15), parameter :: sub = 'def_ncbiop_vars'
    double precision, parameter :: minlat =  -90.0;
    double precision, parameter :: maxlat =   90.0;
    double precision, parameter :: minlon = -180.0;
    double precision, parameter :: maxlon =  180.0;
    double precision :: doublefill(1);			! Set differently for different variables


    !================================================================================
    PROJECT_NAME_dims(2) = N_PROF_dimid;
    PROJECT_NAME_dims(1) = STRING64_dimid;
    ret=nf_def_var(ncid, 'PROJECT_NAME', NF_CHAR,NDIMS2, PROJECT_NAME_dims,PROJECT_NAME_id); call cke(sub,ret, 5);
    ret=nf_put_att_text(ncid, PROJECT_NAME_id, 'long_name', 19, 'Name of the project'); call cke(sub,ret, 6);
    ret=nf_put_att_text(ncid, PROJECT_NAME_id, '_FillValue', 1, ' '); call cke(sub,ret, 7);
    !==================================================================
    PI_NAME_dims(2) = N_PROF_dimid;
    PI_NAME_dims(1) = STRING64_dimid;
    ret=nf_def_var(ncid, 'PI_NAME', NF_CHAR,NDIMS2, PI_NAME_dims,PI_NAME_id); call cke(sub,ret, 8);
    ret=nf_put_att_text(ncid, PI_NAME_id, 'long_name', 34, 'Name of the principal investigator'); call cke(sub,ret, 9);
    ret=nf_put_att_text(ncid, PI_NAME_id, '_FillValue', 1, ' '); call cke(sub,ret,10);
    !==================================================================
    BSTN_PARAMETERS_dims(3) = N_PROF_dimid;		! leftmost dimension		don't ask
    BSTN_PARAMETERS_dims(2) = N_PARAM_dimid;		! middle dimensin		 it just works
    BSTN_PARAMETERS_dims(1) = STRING64_dimid;	! rightmost dimensin		  ! ncbiop PARAMETER names are 64 for B-Argo
    ret=nf_def_var(ncid,'STATION_PARAMETERS',NF_CHAR,NDIMS3,BSTN_PARAMETERS_dims,BSTATION_PARS_id); call cke(sub,ret,33);
    ret=nf_put_att_text(ncid,BSTATION_PARS_id,'long_name',44,'List of available parameters for the station'); call cke(sub,ret,34);
    ret=nf_put_att_text(ncid,BSTATION_PARS_id,'conventions',22,'Argo reference table 3'); call cke(sub,ret,35);
    ret=nf_put_att_text(ncid,BSTATION_PARS_id,'_FillValue',1,' '); call cke(sub,ret,35);
    !==================================================================
    CYCLE_NUMBER_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'CYCLE_NUMBER', NF_INT,NDIMS1, CYCLE_NUMBER_dims,CYCLE_NUMBER_id); call cke(sub,ret,11);
    ret=nf_put_att_text(ncid,CYCLE_NUMBER_id, 'long_name', 18, 'Float cycle number')
    str = '0...N, 0 : launch cycle (if exists), 1 : first complete cycle';
    if(profver3p0) str = '0..N, 0 : launch cycle (if exists), 1 : first complete cycle';
    ret=nf_put_att_text(ncid,CYCLE_NUMBER_id, 'conventions',lt(str),trim(str));
    if(DEBUG)then; write(msg,'("CYCNUM write cycum to prof: ",a)')trim(str); call wrtdbg(sub,msg); endif
    ret=nf_put_att_int(ncid,CYCLE_NUMBER_id, '_FillValue', NF_INT, 1, FILLINT) ! FILLINT int fill value in netcdf_arrays
    !==================================================================
    DIRECTION_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'DIRECTION', NF_CHAR,NDIMS1, DIRECTION_dims,DIRECTION_id); call cke(sub,ret,12);
    ret=nf_put_att_text(ncid,DIRECTION_id,'long_name', 33, 'Direction of the station profiles'); call cke(sub,ret,13);
    ret=nf_put_att_text(ncid,DIRECTION_id,'conventions',45,'A: ascending profiles, D: descending profiles'); call cke(sub,ret,14);
    ret=nf_put_att_text(ncid,DIRECTION_id,'_FillValue', 1, ' '); call cke(sub,ret,15);
    !================================================================================
    DATA_CENTRE_dims(2) = N_PROF_dimid ;
    DATA_CENTRE_dims(1) = STRING2_dimid;
    ret=nf_def_var(ncid, 'DATA_CENTRE', NF_CHAR,NDIMS2, DATA_CENTRE_dims,DATA_CENTRE_id); call cke(sub,ret,16);  ! For ncprof
    ret=nf_put_att_text(ncid, DATA_CENTRE_id, 'long_name', 46, DC_long_name_prof);          call cke(sub,ret,17)
    ret=nf_put_att_text(ncid, DATA_CENTRE_id, 'conventions', 22, 'Argo reference table 4'); call cke(sub,ret,18);
    ret=nf_put_att_text(ncid, DATA_CENTRE_id, '_FillValue', 1, ' '); call cke(sub,ret,19);
    !================================================================================
    DC_REFERENCE_dims(2) = N_PROF_dimid ;
    DC_REFERENCE_dims(1) = STRING32_dimid;
    ret=nf_def_var(ncid, 'DC_REFERENCE', NF_CHAR,NDIMS2, DC_REFERENCE_dims,DC_REFERENCE_id); call cke(sub,ret,20);
    ret=nf_put_att_text(ncid, DC_REFERENCE_id, 'long_name', 40, 'Station unique identifier in data centre'); call cke(sub,ret,21);
    ret=nf_put_att_text(ncid, DC_REFERENCE_id, 'conventions', 22, 'Data centre convention'); call cke(sub,ret,22);
    ret=nf_put_att_text(ncid, DC_REFERENCE_id, '_FillValue', 1, ' '); call cke(sub,ret,23);
    !================================================================================
    DATA_STATE_IND_dims(2) = N_PROF_dimid ;
    DATA_STATE_IND_dims(1) = STRING4_dimid;
    ret=nf_def_var(ncid, 'DATA_STATE_INDICATOR', NF_CHAR,NDIMS2, &
		DATA_STATE_IND_dims,DATA_STATE_IND_id); call cke(sub,ret,24);
    ret=nf_put_att_text(ncid, DATA_STATE_IND_id, 'long_name', 49, DATA_STATE_IND_long_name); call cke(sub,ret,25);
    ret=nf_put_att_text(ncid, DATA_STATE_IND_id, 'conventions', 22, 'Argo reference table 6'); call cke(sub,ret,26);
    ret=nf_put_att_text(ncid, DATA_STATE_IND_id, '_FillValue', 1, ' '); call cke(sub,ret,27);
    !================================================================================
    DATA_MODE_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'DATA_MODE', NF_CHAR,NDIMS1, DATA_MODE_dims,DATA_MODE_id); call cke(sub,ret,28);
    ret=nf_put_att_text(ncid, DATA_MODE_id, 'long_name', 30, 'Delayed mode or real time data'); call cke(sub,ret,29);
    ret=nf_put_att_text(ncid, DATA_MODE_id, 'conventions', 62, DATA_MODE_conventions); call cke(sub,ret,30);
    ret=nf_put_att_text(ncid, DATA_MODE_id, '_FillValue', 1, ' '); call cke(sub,ret,31);
    !==================================================================
    ! 20171016 - added PARAMETER_DATA_MODE for B-Argo profile ncfile
    PARAMETER_DATA_MODE_dims(2) = N_PROF_dimid; PARAMETER_DATA_MODE_dims(1) = N_PARAM_dimid;
    ret=nf_def_var(ncid,'PARAMETER_DATA_MODE',NF_CHAR,NDIMS2,PARAMETER_DATA_MODE_dims,PARAMETER_DATA_MODE_id);call cke(sub,ret,28);
    ret=nf_put_att_text(ncid, PARAMETER_DATA_MODE_id, 'long_name', 30, 'Delayed mode or real time data'); call cke(sub,ret,29);
    ret=nf_put_att_text(ncid, PARAMETER_DATA_MODE_id, 'conventions', 62, DATA_MODE_conventions); call cke(sub,ret,30);
    ret=nf_put_att_text(ncid, PARAMETER_DATA_MODE_id, '_FillValue', 1, ' '); call cke(sub,ret,31);
    !==================================================================
    PLATFORM_TYPE_dims(2) = N_PROF_dimid ;
    PLATFORM_TYPE_dims(1) = STRING32_dimid;
    ret=nf_def_var(ncid, 'PLATFORM_TYPE', NF_CHAR,NDIMS2, PLATFORM_TYPE_dims,PLATFORM_TYPE_id); call cke(sub,ret,42);
    ret=nf_put_att_text(ncid, PLATFORM_TYPE_id, 'long_name', 13, 'Type of float'); call cke(sub,ret,43);
    if(.not. profver3p0)then ! NOTE NO OFFSET FOR if(...
    ret=nf_put_att_text(ncid, PLATFORM_TYPE_id, 'conventions',23, 'Argo reference table 23');
    endif ! profver3p0
    ret=nf_put_att_text(ncid, PLATFORM_TYPE_id, '_FillValue', 1, ' '); call cke(sub,ret,44);
    !==================================================================
    FLOAT_SERIAL_NO_dims(2) = N_PROF_dimid ;
    FLOAT_SERIAL_NO_dims(1) = STRING32_dimid;
    if(profver3p0) FLOAT_SERIAL_NO_dims(1) = STRING16_dimid;
    ret=nf_def_var(ncid, 'FLOAT_SERIAL_NO', NF_CHAR,NDIMS2, FLOAT_SERIAL_NO_dims,FLOAT_SERIAL_NO_id); call cke(sub,ret,23);
    ret=nf_put_att_text(ncid, FLOAT_SERIAL_NO_id, 'long_name', 26, 'Serial number of the float'); call cke(sub,ret,24);
    ret=nf_put_att_text(ncid, FLOAT_SERIAL_NO_id, '_FillValue', 1, ' '); call cke(sub,ret,25);
    !==================================================================
    FIRMWARE_VERSION_dims(2) = N_PROF_dimid ;
    FIRMWARE_VERSION_dims(1) = STRING32_dimid;
    if(profver3p0) FIRMWARE_VERSION_dims(1) = STRING16_dimid;
    ret=nf_def_var(ncid, 'FIRMWARE_VERSION', NF_CHAR,NDIMS2, FIRMWARE_VERSION_dims,FIRMWARE_VERSION_id); call cke(sub,ret,54);
    str = 'Instrument firmware version'
    if(profver3p0) str = 'Instrument version';
    ret=nf_put_att_text(ncid, FIRMWARE_VERSION_id, 'long_name',lt(str),trim(str));
    ret=nf_put_att_text(ncid, FIRMWARE_VERSION_id, '_FillValue', 1, ' '); call cke(sub,ret,52);
    !================================================================================
    WMO_INST_TYPE_dims(2) = N_PROF_dimid ;
    WMO_INST_TYPE_dims(1) = STRING4_dimid;
    ret=nf_def_var(ncid, 'WMO_INST_TYPE', NF_CHAR,NDIMS2, WMO_INST_TYPE_dims,WMO_INST_TYPE_id); call cke(sub,ret,32);
    ret=nf_put_att_text(ncid, WMO_INST_TYPE_id, 'long_name', 21, 'Coded instrument type'); call cke(sub,ret,33);
    ret=nf_put_att_text(ncid, WMO_INST_TYPE_id, 'conventions', 22, 'Argo reference table 8'); call cke(sub,ret,34);
    ret=nf_put_att_text(ncid, WMO_INST_TYPE_id, '_FillValue', 1, ' '); call cke(sub,ret,35);
    !================================================================================
    JULD_dims(1) = N_PROF_dimid
    ret=nf_def_var(ncid, 'JULD', NF_DOUBLE,NDIMS1, JULD_dims,JULD_id); call cke(sub,ret,36);
    ret=nf_put_att_text(ncid, JULD_id, 'long_name', 63, &
                'Julian day (UTC) of the station relative to REFERENCE_DATE_TIME'); call cke(sub,ret,37);
    ret=nf_put_att_text(ncid, JULD_id, 'standard_name', 4, 'time'); call cke(sub,ret,37);
    ret=nf_put_att_text(ncid, JULD_id, 'units', 34,'days since 1950-01-01 00:00:00 UTC'); call cke(sub,ret,38);
    ret=nf_put_att_text(ncid, JULD_id, 'conventions', 56,&
	'Relative julian days with decimal part (as parts of day)'); call cke(sub,ret,39);
    if(.not. profver3p0)then ! Note no offset
    ret=nf_put_att_double(ncid, JULD_id, 'resolution', NF_DOUBLE, 1, JULD_RESOLUTION); call cke(sub,ret,39);
    endif; ! profv3p0
    doublefill(1)=999999.; ret=nf_put_att_double(ncid, JULD_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,40);
    ret=nf_put_att_text(ncid, JULD_id, 'axis', 1, 'T'); call cke(sub,ret,40);
    !================================================================================
    JULD_QC_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'JULD_QC', NF_CHAR,NDIMS1, JULD_QC_dims,JULD_QC_id); call cke(sub,ret,41);
    str = "Quality on date and time";
    if(profver3p0) str = "Quality on Date and Time";
    ret=nf_put_att_text(ncid, JULD_QC_id, 'long_name', lt(str),trim(str)); call cke(sub,ret,42);
    ! OUT if(profver3p0) ret=nf_put_att_text(ncid, JULD_QC_id, 'long_name', 24, 'Quality on Date and Time'); call cke(sub,ret,42);
    ret=nf_put_att_text(ncid, JULD_QC_id, 'conventions', 22, 'Argo reference table 2'); call cke(sub,ret,43);
    ret=nf_put_att_text(ncid, JULD_QC_id, '_FillValue', 1, ' '); call cke(sub,ret,44);
    !================================================================================
    JULD_LOCATION_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'JULD_LOCATION', NF_DOUBLE,NDIMS1, JULD_LOCATION_dims,JULD_LOCATION_id); call cke(sub,ret,45);
    ret=nf_put_att_text(ncid, JULD_LOCATION_id, 'long_name', 64, &
		'Julian day (UTC) of the location relative to REFERENCE_DATE_TIME'); call cke(sub,ret,46);
    ret=nf_put_att_text(ncid, JULD_LOCATION_id, 'units', 34,'days since 1950-01-01 00:00:00 UTC'); call cke(sub,ret,47);
    ret=nf_put_att_text(ncid, JULD_LOCATION_id, 'conventions', 56, &
	'Relative julian days with decimal part (as parts of day)'); call cke(sub,ret,48);
    if(.not. profver3p0)then ! Note no offset
    ret=nf_put_att_double(ncid, JULD_LOCATION_id, 'resolution', NF_DOUBLE, 1, JULD_RESOLUTION); call cke(sub,ret,39);
    endif; ! profv3p0
    doublefill(1)=999999.;
    ret=nf_put_att_double(ncid, JULD_LOCATION_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,49);
    !================================================================================
    LATITUDE_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'LATITUDE', NF_DOUBLE,NDIMS1, LATITUDE_dims,LATITUDE_id); call cke(sub,ret,50);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'long_name', 38,'Latitude of the station, best estimate'); call cke(sub,ret,51);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'standard_name', 8,'latitude'); call cke(sub,ret,52);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'units', 12, 'degree_north'); call cke(sub,ret,53);
    doublefill(1)=99999. ;
    ret=nf_put_att_double(ncid, LATITUDE_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,54);
    ret=nf_put_att_double(ncid, LATITUDE_id, 'valid_min', NF_DOUBLE, 1, minlat); call cke(sub,ret,55);
    ret=nf_put_att_double(ncid, LATITUDE_id, 'valid_max', NF_DOUBLE, 1, maxlat); call cke(sub,ret,56);
    ret=nf_put_att_text(ncid, LATITUDE_id, 'axis', 1, 'Y'); call cke(sub,ret,57);
    !================================================================================
    LONGITUDE_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'LONGITUDE', NF_DOUBLE,NDIMS1, LONGITUDE_dims,LONGITUDE_id); call cke(sub,ret,58);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'long_name', 39,'Longitude of the station, best estimate'); call cke(sub,ret,59);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'standard_name', 9,'longitude'); call cke(sub,ret,60);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'units', 11, 'degree_east'); call cke(sub,ret,61);
    doublefill(1)=99999. ;
    ret=nf_put_att_double(ncid, LONGITUDE_id, '_FillValue', NF_DOUBLE,1,doublefill(1)); call cke(sub,ret,62);
    ret=nf_put_att_double(ncid, LONGITUDE_id, 'valid_min', NF_DOUBLE, 1, minlon); call cke(sub,ret,63);
    ret=nf_put_att_double(ncid, LONGITUDE_id, 'valid_max', NF_DOUBLE, 1, maxlon); call cke(sub,ret,64);
    ret=nf_put_att_text(ncid, LONGITUDE_id, 'axis', 1, 'X'); call cke(sub,ret,65);
    !================================================================================
    POSITION_QC_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid, 'POSITION_QC', NF_CHAR,NDIMS1, POSITION_QC_dims,POSITION_QC_id); call cke(sub,ret,66);
    ret=nf_put_att_text(ncid,POSITION_QC_id,'long_name',44,'Quality on position (latitude and longitude)'); call cke(sub,ret,67);
    ret=nf_put_att_text(ncid, POSITION_QC_id, 'conventions', 22, 'Argo reference table 2'); call cke(sub,ret,68);
    ret=nf_put_att_text(ncid, POSITION_QC_id, '_FillValue', 1, ' '); call cke(sub,ret,69);
    !================================================================================
    POSITIONING_SYSTEM_dims(2) = N_PROF_dimid ;
    POSITIONING_SYSTEM_dims(1) = STRING8_dimid;
    ret=nf_def_var(ncid,'POSITIONING_SYSTEM',NF_CHAR,NDIMS2,POSITIONING_SYSTEM_dims,POSITIONING_SYSTEM_id); call cke(sub,ret,70);
    ret=nf_put_att_text(ncid, POSITIONING_SYSTEM_id, 'long_name', 18, 'Positioning system'); call cke(sub,ret,71);
    ret=nf_put_att_text(ncid, POSITIONING_SYSTEM_id, '_FillValue', 1, ' '); call cke(sub,ret,72);
    !==================================================================
    VERT_SAMPLING_SCHEME_dims(2) = N_PROF_dimid ;
    VERT_SAMPLING_SCHEME_dims(1) = STRING256_dimid;
    ret=nf_def_var(ncid, 'VERTICAL_SAMPLING_SCHEME', NF_CHAR,NDIMS2, VERT_SAMPLING_SCHEME_dims, &
				VERT_SAMPLING_SCHEME_id); call cke(sub,ret,61);
    ret=nf_put_att_text(ncid, VERT_SAMPLING_SCHEME_id, 'long_name', 24, 'Vertical sampling scheme'); call cke(sub,ret,62);
    ret=nf_put_att_text(ncid, VERT_SAMPLING_SCHEME_id, 'conventions', 23, 'Argo reference table 16');
    ret=nf_put_att_text(ncid, VERT_SAMPLING_SCHEME_id, '_FillValue', 1, ' '); call cke(sub,ret,63);
    !================================================================================
    CONFIG_MISSION_NBR_dims(1) = N_PROF_dimid;
    ret=nf_def_var(ncid,'CONFIG_MISSION_NUMBER',NF_INT,NDIMS1,CONFIG_MISSION_NBR_dims,CONFIG_MISSION_NBR_id); call cke(sub,ret,11);
    str = "Unique number denoting the missions performed by the float";
    if(profver3p0) str = "Float's mission number for each profile";;
    ret=nf_put_att_text(ncid,CONFIG_MISSION_NBR_id, 'long_name',lt(str),trim(str));
    str = "1...N, 1 : first complete mission";
    if(profver3p0) str = "0..N, 0 : launch mission (if exists), 1 : first complete mission";
    ret=nf_put_att_text(ncid,CONFIG_MISSION_NBR_id,'conventions',lt(str),trim(str));
    ret=nf_put_att_int(ncid,CONFIG_MISSION_NBR_id, '_FillValue', NF_INT, 1, FILLINT) ! FILLINT int fill value in netcdf_arrays
    !==================================================================

end subroutine def_ncbiop_vars

  !====================================================================
subroutine put_ncbiop_vars(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;
    integer :: ihst, ipar, ical ! Counters for the dimensions

    integer :: start3(3), count3(3);	! corners for call for BSTN_PARAMETERS

    character(len=4) :: c4		! Array messge transfer
    character(len=64) :: c64;           ! if-else-then storage

    character(len=15), parameter :: sub = 'put_ncbiop_vars'

    !================================================================================
    do i=1,N_PROF; PROJECT_NAME(i) = chr_PROJECT_NAME; enddo
     ret=nf_put_var_text(ncid, PROJECT_NAME_id, PROJECT_NAME); call cke(sub,ret,74);
    !================================================================================
    do i=1,N_PROF; PI_NAME(i) = var_PI_NAME; enddo
     ret=nf_put_var_text(ncid, PI_NAME_id, PI_NAME); call cke(sub,ret,75);
    !================================================================================
    ! Count the station parameters (also used for the SCICAL PARAMETERs
    ! First, load the "real" parameters
    do iprf=1,nbprf;
	ipar=1;                        BSTN_PARAMETERS(ipar) = bparpres; BPARAMETERS(ipar,iprf) = bparpres; ! All profiles get PRES
	ipar=2; if(hasDOXY(iprf))then; BSTN_PARAMETERS(ipar) = bpardoxy; BPARAMETERS(ipar,iprf) = bpardoxy; endif

        if(hasTEMP_DOXY(iprf))then;
	    ipar=ipar+1; hasBDAT(ipar,iprf) = .true.;
	    BSTN_PARAMETERS(ipar)=BPARNAME(1); BPARAMETERS(ipar,iprf)=BPARNAME(1);
	endif
        if(hasFREQ_DOXY(iprf))then;
	    ipar=ipar+1; hasBDAT(ipar,iprf) = .true.;
	    BSTN_PARAMETERS(ipar)=BPARNAME(2); BPARAMETERS(ipar,iprf)=BPARNAME(2);
	endif
        if(hasBPHA_DOXY(iprf))then;
	    ipar=ipar+1; hasBDAT(ipar,iprf) = .true.;
	    BSTN_PARAMETERS(ipar)=BPARNAME(3); BPARAMETERS(ipar,iprf)=BPARNAME(3);
	endif
        if(hasTPHA_DOXY(iprf))then;
	    ipar=ipar+1; hasBDAT(ipar,iprf) = .true.;
	    BSTN_PARAMETERS(ipar)=BPARNAME(4); BPARAMETERS(ipar,iprf)=BPARNAME(4);
	endif
        if(hasRPHA_DOXY(iprf))then;
	    ipar=ipar+1; hasBDAT(ipar,iprf) = .true.;
	    BSTN_PARAMETERS(ipar)=BPARNAME(5); BPARAMETERS(ipar,iprf)=BPARNAME(5);
	endif

	start3(1) = 1;  start3(2) = 1;      start3(3) = iprf;
	count3(1) = 64; count3(2) = nbparout; count3(3) = 1;
	 ret=nf_put_vara_text(ncid, BSTATION_PARS_id,start3, count3,BSTN_PARAMETERS); call cke(sub,ret,36);

    enddo

    ! STATION_PARAMETERS Special case - 1 inst-type has 4 profiles, normal primary, DOXY in prf2, near-surf data in prf3 34
    if(nprf > nbprf)then;
        ! For the BR file, it is reuired to have PRES as a staion parameter for all profiles
        do iprf=nbprf+1,nprf ! iprf=1,nbprf;
            ipar=1; BSTN_PARAMETERS(ipar) = bparpres; BPARAMETERS(ipar,iprf) = bparpres; ! All profiles get PRES
	    start3(1) = 1;  start3(2) = 1;      start3(3) = iprf;
	    count3(1) = 64; count3(2) = 1; count3(3) = 1; ! Only load the FIRST parameter, PRES
	     ret=nf_put_vara_text(ncid, BSTATION_PARS_id,start3, count3,BSTN_PARAMETERS); call cke(sub,ret,36);
        enddo
    endif


    do i=1,N_PROF; CYCLE_NUMBER(i) = thiscyc; enddo;
     ret=nf_put_var_int(ncid, CYCLE_NUMBER_id, CYCLE_NUMBER); call cke(sub,ret,76);
    !================================================================================
    do i=1,N_PROF; DIRECTION(i) = chr_DIRECTION; enddo
     ret=nf_put_var_text(ncid, DIRECTION_id, DIRECTION); call cke(sub,ret,77);
    !================================================================================
    do i=1,N_PROF; DATA_CENTRE(i) = chr_DATA_CENTRE; enddo
     ret=nf_put_var_text(ncid, DATA_CENTRE_id, DATA_CENTRE); call cke(sub,ret,78);
    !================================================================================
    do i=1,N_PROF; DC_REFERENCE(i) = chr_DC_REFERENCE; enddo
     ret=nf_put_var_text(ncid, DC_REFERENCE_id, DC_REFERENCE); call cke(sub,ret,79);
    !================================================================================
    do i=1,N_PROF; DATA_STATE_INDICATOR(i) = chr_DATA_STATE_INDICATOR; enddo
     ret=nf_put_var_text(ncid, DATA_STATE_IND_id, DATA_STATE_INDICATOR); call cke(sub,ret,80);
    !================================================================================
    do i=1,N_PROF; DATA_MODE(i) = 'R'; enddo; ! ncbiop - mode must be R chr_DATA_MODE; enddo
     ret=nf_put_var_text(ncid, DATA_MODE_id, DATA_MODE); call cke(sub,ret,81);
    !================================================================================
    ! 20171016 lod the new PARAMETER_DATA_MODE biop variable
    ! FIX THIS KLUGE
    ! ??? PARAMETER_DATA_MODE(MAXPRF,MAXBPAR);
    do iprf=1,MAXPRF; do ipar=1,MAXBPAR; PARAMETER_DATA_MODE(iprf,Ipar) = 'R'; enddo; enddo;

     ret=nf_put_var_text(ncid, PARAMETER_DATA_MODE_id, PARAMETER_DATA_MODE); call cke(sub,ret,81);
    !================================================================================
    do i=1,N_PROF; PLATFORM_TYPE(i) = chr_PLATFORM_TYPE; enddo
     ret=nf_put_var_text(ncid, PLATFORM_TYPE_id, PLATFORM_TYPE); call cke(sub,ret,45);
    !================================================================================
    if(profver3p0)then;
	do i=1,N_PROF; FLOAT_SERIAL_NO_3p0(i) = var_FLOAT_SERIAL_NO; enddo;
	 ret=nf_put_var_text(ncid, FLOAT_SERIAL_NO_id, FLOAT_SERIAL_NO_3p0); call cke(sub,ret,26);
    else; ! profver3p0
	do i=1,N_PROF; FLOAT_SERIAL_NO(i) = var_FLOAT_SERIAL_NO; enddo
	 ret=nf_put_var_text(ncid, FLOAT_SERIAL_NO_id, FLOAT_SERIAL_NO); call cke(sub,ret,26);
    endif ! profver3p0 - NOTE the inconsistent indentation for profver3p1
    !================================================================================
    do i=1,N_PROF; FIRMWARE_VERSION(i) = var_FIRMWARE_VERSION; enddo
    if(profver3p0)then;
	do i=1,N_PROF; FIRMWARE_VERSION_3p0(i) = var_FIRMWARE_VERSION; enddo
	 ret=nf_put_var_text(ncid, FIRMWARE_VERSION_id, FIRMWARE_VERSION_3p0); call cke(sub,ret,53);
    else ! profver3p0
     ret=nf_put_var_text(ncid, FIRMWARE_VERSION_id, FIRMWARE_VERSION); call cke(sub,ret,53);
    endif ! profver3p0 - NOTE the inconsistent indentation for profver3p1
    !================================================================================
    do i=1,N_PROF; WMO_INST_TYPE(i) = var_WMO_INST_TYPE; enddo
     ret=nf_put_var_text(ncid, WMO_INST_TYPE_id, WMO_INST_TYPE); call cke(sub,ret,82);
    !================================================================================
    do i=1,N_PROF; JULD(i) = prof_JULD; enddo
     ret=nf_put_var_double(ncid, JULD_id, JULD); call cke(sub,ret,83);
    !================================================================================
    do i=1,N_PROF; JULD_QC(i) = prof_JULD_QC; enddo
     ret=nf_put_var_text(ncid, JULD_QC_id, JULD_QC); call cke(sub,ret,84);
    !================================================================================
    do i=1,N_PROF; JULD_LOCATION(i) = prof_JULD_LOCATION; enddo
     ret=nf_put_var_double(ncid, JULD_LOCATION_id, JULD_LOCATION); call cke(sub,ret,85);
    !================================================================================
    do i=1,N_PROF; LATITUDE(i) = prof_LATITUDE; enddo
     ret=nf_put_var_double(ncid, LATITUDE_id, LATITUDE); call cke(sub,ret,86);
    !================================================================================
    do i=1,N_PROF; LONGITUDE(i) = prof_LONGITUDE; enddo
     ret=nf_put_var_double(ncid, LONGITUDE_id, LONGITUDE); call cke(sub,ret,87);
    !================================================================================
    do i=1,N_PROF; POSITION_QC(i) = prof_POSITION_QC; enddo
     ret=nf_put_var_text(ncid, POSITION_QC_id, POSITION_QC); call cke(sub,ret,88);
    !================================================================================
    do i=1,N_PROF; POSITIONING_SYSTEM(i) = chr_POSITIONING_SYSTEM; enddo
     ret=nf_put_var_text(ncid, POSITIONING_SYSTEM_id, POSITIONING_SYSTEM); call cke(sub,ret,89);
    !================================================================================
    do i=1,N_PROF; VERT_SAMPLING_SCHEME(i) = chr_VERT_SAMPLING_SCHEME(i); enddo
     ret=nf_put_var_text(ncid, VERT_SAMPLING_SCHEME_id, VERT_SAMPLING_SCHEME); call cke(sub,ret,64);
    !================================================================================
    do i=1,N_PROF; CONFIG_MISSION_NUMBER(i) = int_CONFIG_MISSION_NUMBER; enddo
     ret=nf_put_var_int(ncid, CONFIG_MISSION_NBR_id, CONFIG_MISSION_NUMBER); call cke(sub,ret,76);
    !================================================================================

end subroutine put_ncbiop_vars

  !====================================================================
subroutine def_ncbiop_params(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;
    integer :: ipar;
    integer :: TMP_id		! tmp id for PRES (only need 1, not 5)
    integer :: TMP_ids(5);	! PAR PAR_QC PAR_ADJ PAR_ADJ_QC PAR_ERR

    character(len=48) :: nam

    character(len=16), parameter :: sub = 'def_ncbiop_parms'

    N_PARAM = nparam;

    if(DBLEV>0)then;write(msg,*)'In ',sub,' hasNCPAR: ',hasNCPAR; call wrtdbg(sub,msg); endif

! Before we define the dat parameters, w need to define the PROFILE_<param>_QC FIRST, before the parameters
! Note that we do NOT need this for PRES, as the QC'd data is limited to the CoreARGO file

! We ALWAYS have DOXY

call def_ncbiop_profile_qc(sub,ncid,'DOXY',DOXY_profqc_id);

ipar = 0; ! This counter is for the BPAR ids, PRES & DOXY omitted
if(hasTEMP_DOXY(0))then
    ipar = ipar + 1; call def_ncbiop_profile_qc(sub,ncid,'TEMP_DOXY',TMP_id); BBPAR_profqc_ids(ipar) = TMP_id;
endif
if(hasFREQ_DOXY(0))then
    ipar = ipar + 1; call def_ncbiop_profile_qc(sub,ncid,'FREQUENCY_DOXY',TMP_id); BBPAR_profqc_ids(ipar) = TMP_id;
endif
if(hasBPHA_DOXY(0))then
    ipar = ipar + 1; call def_ncbiop_profile_qc(sub,ncid,'BPHASE_DOXY',TMP_id); BBPAR_profqc_ids(ipar) = TMP_id;
endif
if(hasTPHA_DOXY(0))then
    ipar = ipar + 1; call def_ncbiop_profile_qc(sub,ncid,'TPHASE_DOXY',TMP_id); BBPAR_profqc_ids(ipar) = TMP_id;
endif
if(hasRPHA_DOXY(0))then
    ipar = ipar + 1; call def_ncbiop_profile_qc(sub,ncid,'RPHASE_DOXY',TMP_id); BBPAR_profqc_ids(ipar) = TMP_id;
endif

! Now add the actual parmeter arrays

call def_ncbiop_param_pres(sub,ncid,'PRES',TMP_id);
BPRES_id = TMP_id;

call def_ncbiop_param_doxy(sub,ncid,'DOXY',TMP_ids);
do inca=1,5; BDOXY_ids(inca) = TMP_ids(inca); enddo

ipar = 0;
if(hasTEMP_DOXY(0))then
    ipar = 1;
    call def_ncbiop_param(sub,ncid,ipar,BPARNAME(ipar),TMP_ids);do i=1,5; TEMP_DOXY_ids(i) = TMP_ids(i); enddo
endif
if(hasFREQ_DOXY(0))then
    ipar = 2;
    call def_ncbiop_param(sub,ncid,ipar,BPARNAME(ipar),TMP_ids);do i=1,5; FREQ_DOXY_ids(i) = TMP_ids(i); enddo
endif
if(hasBPHA_DOXY(0))then
    ipar = 3;
    call def_ncbiop_param(sub,ncid,ipar,BPARNAME(ipar),TMP_ids);do i=1,5; BPHA_DOXY_ids(i) = TMP_ids(i); enddo
endif
if(hasTPHA_DOXY(0))then
    ipar = 4;
    call def_ncbiop_param(sub,ncid,ipar,BPARNAME(ipar),TMP_ids);do i=1,5; TPHA_DOXY_ids(i) = TMP_ids(i); enddo
endif
if(hasRPHA_DOXY(0))then
    ipar = 5;
    call def_ncbiop_param(sub,ncid,ipar,BPARNAME(ipar),TMP_ids);do i=1,5; RPHA_DOXY_ids(i) = TMP_ids(i); enddo
endif


    if(DBLEV>0)then; write(msg,'("<< Leave >>")'); call wrtdbg(sub,msg); endif

    !==================================================================

end subroutine def_ncbiop_params

  !====================================================================
subroutine put_ncbiop_params(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer :: loc = 0;
    integer :: ipar
    integer :: TMP_ids(5);      ! PAR PAR_QC PAR_ADJ PAR_ADJ_QC PAR_ERR

    integer :: iloc, ipin

    character(len=16), parameter :: sub = 'put_ncbiop_parms'

    if(DBLEV>0)then; write(msg,'("<< Enter ",a," >>")')sub; call wrtdbg(sub,msg); endif

    ! NOTE do NOT try to fill any prof_<>_qc EXCEPT DOXY  it is the only parameter that is actually qc'd
    ! If we do nothing for the additional parameters, they will just remain as fill value -the cooorect  value

    ! Next we fill up the parameter arrays

    call put_ncbiop_profile_qc(sub,ncid,'DOXY',DOXY_profqc_id);

    ! Next we fill up the parameter arrays

!OLD note offset below

do ilev=1,nlev; FDATA(ilev) = FILLNC; FFLAG(ilev) = FILLCHAR; FNOTD = '0'; enddo

! put PRES into ncbip file

TMP_ids(1) = BPRES_id;
ipar = 1; ! PRES is always first
do iprf=1,nprf
    ! From data_arrays:  regPRES(MAXPRF) = (/p1r,p2r,p3r,p4r/);
    iloc = regPRES(iprf);
    do ilev=1,nlev;
	if(profdata(iloc,ilev) == FILL)then
	    BDATA(ilev) = FILLNC; ! redundent, aoids naked ';'
	else
	    BDATA(ilev) = profdata(iloc,ilev)
	endif
    enddo
    call put_ncbiop_param(sub,ncid,ipar,iprf,'PRES',TMP_ids);
enddo

! put DOXY into ncbip file

do i=1,5; TMP_ids(i) = BDOXY_ids(i); enddo
ipar = 2; ! DOXY is always second
do iprf=1,nprf
    iloc = 0; ! We do not have data hin this profile
    if(iprf == 1)then
	iloc = do1;
    elseif(iprf == 2)then
	if(hasDOXY(2))then
	    iloc = do2;
	elseif(hasDOXY(3))then
	    iloc = doa;
	endif
    endif
    if(iloc > 0)then ! We DO have doxy data in thisprofile
	do ilev=1,nlev
	    if(profflag(iloc,ilev) == MISSING)then
		BDATA(ilev) = FILLNC;
		BFLAG(ilev) = FILLCHAR; ! for DOXY
	    else
		BDATA(ilev) = profdata(iloc,ilev);
		BFLAG(ilev) = ncflag(profflag(iloc,ilev)) ! for DOXYG
	    endif
	enddo
	call put_ncbiop_param(sub,ncid,ipar,iprf,'DOXY',TMP_ids);
    endif ! We DO have data
enddo
ipar = 2;
if(DEBUG) write(DBG,'("===============================================================")')
if(hasTEMP_DOXY(0))then
    ipar = ipar + 1;
    do iprf=1,nbprf;
	do ilev=1,nlev
	    if(profbdata(1,iprf,ilev) == FILL)then; BDATA(ilev) = FILLNC; else; BDATA(ilev) = profbdata(1,iprf,ilev); endif
	enddo
	call put_ncbiop_param(sub,ncid,ipar,iprf,'TEMP_DOXY',TEMP_DOXY_ids);
    enddo
endif
if(hasFREQ_DOXY(0))then
    do iprf=1,nbprf;
        do ilev=1,nlev
            if(profbdata(2,iprf,ilev) == FILL)then;
		BDATA(ilev) = FILLNC; BFLAG(ilev) = ' ';
	    else;
		BDATA(ilev) = profbdata(2,iprf,ilev); BFLAG(ilev) = '0';
	    endif
        enddo
        call put_ncbiop_param(sub,ncid,ipar,iprf,'FREQ_DOXY',FREQ_DOXY_ids);
    enddo
endif
if(hasBPHA_DOXY(0))then
    ipar = ipar + 1;
    do iprf=1,nbprf;
	do ilev=1,nlev
	    if(profbdata(3,iprf,ilev) == FILL)then; BDATA(ilev) = FILLNC; else; BDATA(ilev) = profbdata(3,iprf,ilev); endif
	enddo
	call put_ncbiop_param(sub,ncid,ipar,iprf,'BPHA_DOXY',BPHA_DOXY_ids);
    enddo
endif
if(hasTPHA_DOXY(0))then
    ipar = ipar + 1;
    do iprf=1,nbprf;
	do ilev=1,nlev
	    if(profbdata(4,iprf,ilev) == FILL)then; BDATA(ilev) = FILLNC; else; BDATA(ilev) = profbdata(4,iprf,ilev); endif
	enddo
	call put_ncbiop_param(sub,ncid,ipar,iprf,'TPHA_DOXY',TPHA_DOXY_ids);
    enddo
endif
if(hasRPHA_DOXY(0))then
    ipar = ipar + 1;
    do iprf=1,nbprf;
	do ilev=1,nlev
	    if(profbdata(5,iprf,ilev) == FILL)then; BDATA(ilev) = FILLNC; else; BDATA(ilev) = profbdata(5,iprf,ilev); endif
	enddo
	call put_ncbiop_param(sub,ncid,ipar,iprf,'RPHA_DOXY',RPHA_DOXY_ids);
    enddo
endif
    !==================================================================

end subroutine put_ncbiop_params

  !====================================================================
subroutine def_ncbiop_profile_qc(caller,ncid,PAR,PAR_id);
    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid 	! LNetCDF file id
    character(*), intent( in )      :: PAR	! PARAMETER to define
    integer, intent(out )          :: PAR_id;

    integer :: DATA_dim;
    character(len=48) :: nam
    integer :: lst;

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    character(len=18), parameter   :: sub = 'def_ncbiop_prof_qc';

    DATA_dim = N_PROF_dimid;

    nam="PROFILE_" // trim(PAR)//"_QC";  ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS1,DATA_dim,PAR_id); call cke(sub,ret,1);

    nam="Global quality flag of " // trim(PAR) // " profile"; lst = len(trim(nam))
     ret=nf_put_att_text(ncid, PAR_id, 'long_name', lst, nam); call cke(sub,ret, 2);
     ret=nf_put_att_text(ncid, PAR_id, 'conventions', 23, 'Argo reference table 2a")'); call cke(sub,ret, 3);
     ret=nf_put_att_text(ncid, PAR_id, '_FillValue', 1, ' '); call cke(sub,ret, 4);

    20 continue

end subroutine def_ncbiop_profile_qc

  !====================================================================
subroutine put_ncbiop_profile_qc(caller,ncid,PAR,PAR_id);
    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid      ! LNetCDF file id
    character(*), intent( in )     :: PAR      ! PARAMETER to load (only used for annnotation)
    integer, intent( in )          :: PAR_id;

    integer :: iret
    integer :: ret = 0;
    integer :: loc
    integer          :: ipar
    character(len=1) :: grade(MAXPAR) = (/ "x", "x", "x", "x", "x" /);
    character(len=18), parameter   :: sub = 'put_ncp_profile_qc';

    if(DBLEV>0)then; write(msg,'("PR_QC: Put profile_qc for ",a)')PAR; call wrtdbg(sub,msg); endif

    ipar = 4;
    do iprf=1,nprf
	grade(iprf) = prof_doxy_qc(iprf)
	if(grade(iprf) == 'M') grade(iprf) = " "; ! nc file missing vlue flag
	if(grade(iprf) == 'Z') grade(iprf) = " "; ! nc file missing vlue flag
    enddo

    ! Note that these refer to the DATA order (ie, profdata, etc.) notthe NCPAR order

    if(DBLEV>0)then;
	write(msg,*)'NC_PRQC: put_ncbiop_profile_qc PAR: ',PAR,' PAR_id: ',PAR_id,' grade: ',grade;
	call wrtdbg(sub,msg);
    endif
     ret=nf_put_var_text(ncid, PAR_id, grade);    call cke(sub,ret,1);

end subroutine put_ncbiop_profile_qc

  !====================================================================
subroutine def_ncbiop_param_pres(caller,ncid,PAR,PAR_id);
character(*), intent( in )     :: caller
integer, intent( in )          :: ncid      ! LNetCDF file id
character(*), intent( in )      :: PAR      ! PARAMETER to define
integer, intent(out )          :: PAR_id;   ! The PRES parameter needs only one par_id

integer :: DATA_dims(NDIMS2);
character(len=48) :: nam
                             !12345678901234567890123456789012345678901
character(len=64) :: lname = "Sea water pressure, equals 0 at sea-level"; integer :: llname = 41;
character(len=64) :: sname = "sea_water_pressure";                        integer :: lsname = 18;
real              :: FillV = FILLNC;
character(len=8)  :: units = "decibar";                                   integer :: lunits =  7;
real              :: v_min = 0.0;
real              :: v_max = 12000.0;
character(len=8)  :: C_fmt = "%7.1f";                                     integer :: lC_fmt = 5;
character(len=8)  :: F_fmt = "F7.1";                                      integer :: lF_fmt = 4;
real              :: resln = 0.1;

integer :: iret
integer :: ret = 0;
integer :: loc
character(len=21), parameter   :: sub = 'def_ncbiop_param_pres';

DATA_dims(2)         = N_PROF_dimid;  DATA_dims(1)         = N_LEVELS_dimid;

    nam=trim(PAR);
     ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_id); call cke(sub,ret,1);
     ret=nf_put_att_text(ncid, PAR_id, 'long_name',     llname,     lname);
     ret=nf_put_att_text(ncid, PAR_id, 'standard_name', lsname,     sname);
     ret=nf_put_att_real(ncid, PAR_id, '_FillValue',    NF_REAL, 1, FillV);
     ret=nf_put_att_text(ncid, PAR_id, 'units',         lunits,     units);
     ret=nf_put_att_real(ncid, PAR_id, 'valid_min',     NF_REAL, 1, v_min);
     ret=nf_put_att_real(ncid, PAR_id, 'valid_max',     NF_REAL, 1, v_max);
     ret=nf_put_att_text(ncid, PAR_id, 'C_format',      lC_fmt,     C_fmt);
     ret=nf_put_att_text(ncid, PAR_id, 'FORTRAN_format',lF_fmt,     F_fmt);
     ret=nf_put_att_real(ncid, PAR_id, 'resolution',    NF_REAL, 1, resln);
     ret=nf_put_att_text(ncid, PAR_id, 'axis',1, 'Z')


end subroutine def_ncbiop_param_pres

  !====================================================================
  ! SCIENTIFIC_CALIB defs
  !====================================================================
subroutine def_ncbiop_calhis(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;

    character(len=64) :: long_name;
    character(len=70) :: conventns;
    integer :: nn, nc, nu;  ! length of long_name  & conventions

    character(len=16), parameter :: sub = 'def_ncbiop_calhis'

    if(DBLEV==37)then; write(msg,'("LEV37 Enter sub ",a)')sub; call wrtdbg(sub,msg); endif

    !===================================================================
    SCICAL_dims(4) = N_PROF_dimid;
    SCICAL_dims(3) = N_CALIB_dimid;
    SCICAL_dims(2) = N_PARAM_dimid;
    !-------------------------------------------------------------------
    SCICAL_dims(1) = STRING64_dimid; ! ncbiop PARAMETER names are 64 for B-Argo
    !-------------------------------------------------------------------
    if(DBLEV==37)then; write(msg,*)'LEV37 BEFORE def, SCICAL_ids: ',SCICAL_ids; call wrtdbg(sub,msg); endif
     ret=nf_def_var(ncid, 'PARAMETER',                    NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(1)); call cke(sub,ret,11);
    if(DBLEV==37)then; write(msg,*)'LEV37 AFTER  def, SCICAL_ids: ',SCICAL_ids; call wrtdbg(sub,msg); endif
    long_name = "List of parameters with calibration information";                       nn = len(trim(long_name))
    conventns = "Argo reference table 3";                                                nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, SCICAL_ids(1), 'long_name',   nn, long_name);           call cke(sub,ret,12);
     ret=nf_put_att_text(ncid, SCICAL_ids(1), 'conventions', nc, conventns);           call cke(sub,ret,13);
     ret=nf_put_att_text(ncid, SCICAL_ids(1), '_FillValue', 1, ' ');                   call cke(sub,ret,14);
    !-------------------------------------------------------------------
    SCICAL_dims(1) = STRING256_dimid;
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_EQUATION',    NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(2)); call cke(sub,ret,21);
    long_name = "Calibration equation for this parameter";                               nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(2), 'long_name',   nn, long_name);           call cke(sub,ret,22);
     ret=nf_put_att_text(ncid, SCICAL_ids(2), '_FillValue', 1, ' ');                   call cke(sub,ret,23);
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_COEFFICIENT', NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(3)); call cke(sub,ret,31);
    long_name = "Calibration coefficients for this equation";                             nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(3), 'long_name',   nn, long_name);           call cke(sub,ret,32);
     ret=nf_put_att_text(ncid, SCICAL_ids(3), '_FillValue', 1, ' ');                   call cke(sub,ret,33);
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_COMMENT',     NF_CHAR,NDIMS4, SCICAL_dims, SCICAL_ids(4)); call cke(sub,ret,41);
    long_name = "Comment applying to this parameter calibration";                        nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(4), 'long_name',   nn, long_name);           call cke(sub,ret,42);
     ret=nf_put_att_text(ncid, SCICAL_ids(4), '_FillValue', 1, ' ');                   call cke(sub,ret,43);
    !-------------------------------------------------------------------
    SCICAL_dims(1) = DATE_TIME_dimid;
    !-------------------------------------------------------------------
     ret=nf_def_var(ncid, 'SCIENTIFIC_CALIB_DATE',        NF_CHAR,NDIMS4, SCICAL_dims,SCICAL_ids(5)); call cke(sub,ret,51);
    long_name = "Date of calibration"; nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, SCICAL_ids(5), 'long_name',   nn, long_name);           call cke(sub,ret,52);
     ret=nf_put_att_text(ncid, SCICAL_ids(5), 'conventions',  14,'YYYYMMDDHHMISS');
     ret=nf_put_att_text(ncid, SCICAL_ids(5), '_FillValue', 1, ' ');                   call cke(sub,ret,53);

    !================================================================================
    ! HISTORY defs
    !================================================================================
    ! We need both HIST_dim2 and HIST_DIM3 - there are 12 history variables - three of them
    ! are REALs, and have TWO dimentions:
    !    float HISTORY_START_PRES      (N_HISTORY, N_PROF) ;
    !    float HISTORY_STOP_PRES       (N_HISTORY, N_PROF) ;
    !    float HISTORY_PREVIOUS_VALUE  (N_HISTORY, N_PROF) ;
    ! and the rest are character, and have a third dimensiont (size of the char string):
    !    char HISTORY_INSTITUTION      (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_STEP             (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_SOFTWARE         (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_SOFTWARE_RELEASE (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_REFERENCE        (N_HISTORY, N_PROF, STRING64) ;
    !    char HISTORY_DATE             (N_HISTORY, N_PROF, DATE_TIME) ;
    !    char HISTORY_ACTION           (N_HISTORY, N_PROF, STRING4) ;
    !    char HISTORY_PARAMETER        (N_HISTORY, N_PROF, STRING64) ; ! ncbiop PARAMETER names are 64 for B-Argo
    !    char HISTORY_QCTEST           (N_HISTORY, N_PROF, STRING16) ;

    HIST_dim3(3) = N_HISTORY_dimid ! for all the 3D variables (STRING*nn is the first element)
    HIST_dim3(2) = N_PROF_dimid    ! for all the 3D variables
    HIST_dim2(2) = N_HISTORY_dimid ! for all the 2D variables
    HIST_dim2(1) = N_PROF_dimid    ! for all the 2D variables
    ! 01c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_INSTITUTION      (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_INSTITUTION',     NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 1));    call cke(sub,ret,1010);
    long_name = "Institution which performed action";                                               nn = len(trim(long_name))
    conventns = "Argo reference table 4";                                                           nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 1), 'long_name',   nn, long_name);                       call cke(sub,ret,1011);
     ret=nf_put_att_text(ncid, HIST_ids( 1), 'conventions', nc, conventns);                       call cke(sub,ret,1012);
     ret=nf_put_att_text(ncid, HIST_ids( 1), '_FillValue',   1, ' ');                             call cke(sub,ret,1013);
    ! 02c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_STEP             (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_STEP',            NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 2));    call cke(sub,ret,1020);
    long_name = "Step in data processing";                                                          nn = len(trim(long_name))
    conventns = "Argo reference table 12";                                                          nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 2), 'long_name',   nn, long_name);                       call cke(sub,ret,1021);
     ret=nf_put_att_text(ncid, HIST_ids( 2), 'conventions', nc, conventns);                       call cke(sub,ret,1022);
     ret=nf_put_att_text(ncid, HIST_ids( 2), '_FillValue',   1, ' ');                             call cke(sub,ret,1023);
    ! 03c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_SOFTWARE         (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_SOFTWARE',        NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 3));    call cke(sub,ret,1030);
    long_name = "Name of software which performed action";                                          nn = len(trim(long_name))
    conventns = "Institution dependent";                                                            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 3), 'long_name',   nn, long_name);                       call cke(sub,ret,1031);
     ret=nf_put_att_text(ncid, HIST_ids( 3), 'conventions', nc, conventns);                       call cke(sub,ret,1032);
     ret=nf_put_att_text(ncid, HIST_ids( 3), '_FillValue',   1, ' ');                             call cke(sub,ret,1033);
    ! 04c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_SOFTWARE_RELEASE (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_SOFTWARE_RELEASE',NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 4));    call cke(sub,ret,1040);
    long_name = "Version/release of software which performed action";                               nn = len(trim(long_name))
    conventns = "Institution dependent";                                                            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 4), 'long_name',   nn, long_name);                       call cke(sub,ret,1041);
     ret=nf_put_att_text(ncid, HIST_ids( 4), 'conventions', nc, conventns);                       call cke(sub,ret,1042);
     ret=nf_put_att_text(ncid, HIST_ids( 4), '_FillValue',   1, ' ');                             call cke(sub,ret,1043);
    ! 05c-----------------------------------------------------------------
    HIST_dim3(1) = STRING64_dimid; ! char HISTORY_REFERENCE        (N_HISTORY, STRING64);
     ret=nf_def_var(ncid, 'HISTORY_REFERENCE',       NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 5));    call cke(sub,ret,1050);
    long_name = "Reference of database";                                                            nn = len(trim(long_name))
    conventns = "Institution dependent";                                                            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 5), 'long_name',   nn, long_name);                       call cke(sub,ret,1051);
     ret=nf_put_att_text(ncid, HIST_ids( 5), 'conventions', nc, conventns);                       call cke(sub,ret,1052);
     ret=nf_put_att_text(ncid, HIST_ids( 5), '_FillValue',   1, ' ');                             call cke(sub,ret,1053);
    ! 06d-----------------------------------------------------------------
    HIST_dim3(1) = DATE_TIME_dimid ! char HISTORY_DATE             (N_HISTORY, DATE_TIME);
     ret=nf_def_var(ncid, 'HISTORY_DATE',            NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 6));    call cke(sub,ret,1060);
    long_name = "Date the history record was created";                                              nn = len(trim(long_name))
    conventns = "YYYYMMDDHHMISS";                                                                   nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 6), 'long_name',   nn, long_name);                       call cke(sub,ret,1061);
     ret=nf_put_att_text(ncid, HIST_ids( 6), 'conventions', nc, conventns);                       call cke(sub,ret,1062);
     ret=nf_put_att_text(ncid, HIST_ids( 6), '_FillValue',   1, ' ');                             call cke(sub,ret,1063);
    ! 07c-----------------------------------------------------------------
    HIST_dim3(1) = STRING4_dimid;  ! char HISTORY_ACTION           (N_HISTORY, STRING4);
     ret=nf_def_var(ncid, 'HISTORY_ACTION',          NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 7));    call cke(sub,ret,1070);
    long_name = "Action performed on data";                                                         nn = len(trim(long_name))
    conventns = "Argo reference table 7";                                                           nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 7), 'long_name',   nn, long_name);                       call cke(sub,ret,1071);
     ret=nf_put_att_text(ncid, HIST_ids( 7), 'conventions', nc, conventns);                       call cke(sub,ret,1072);
     ret=nf_put_att_text(ncid, HIST_ids( 7), '_FillValue',   1, ' ');                             call cke(sub,ret,1073);
    ! 08c-----------------------------------------------------------------
    HIST_dim3(1) = STRING64_dimid; ! char HISTORY_PARAMETER        (N_HISTORY, STRING4); ! ncbiop PARAMETER names are 64 for B-Argo
     ret=nf_def_var(ncid, 'HISTORY_PARAMETER',       NF_CHAR,NDIMS3, HIST_dim3, HIST_ids( 8));    call cke(sub,ret,1080);
    long_name = "Station parameter action is performed on";                                         nn = len(trim(long_name))
    conventns = "Argo reference table 3";                                                           nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids( 8), 'long_name',   nn, long_name);                       call cke(sub,ret,1081);
     ret=nf_put_att_text(ncid, HIST_ids( 8), 'conventions', nc, conventns);                       call cke(sub,ret,1082);
     ret=nf_put_att_text(ncid, HIST_ids( 8), '_FillValue',   1, ' ');                             call cke(sub,ret,1083);
    ! 09c-----------------------------------------------------------------
     ret=nf_def_var(ncid, 'HISTORY_START_PRES',     NF_REAL, NDIMS2, HIST_dim2, HIST_ids( 9));    call cke(sub,ret,1110);
    long_name = "Start pressure action applied on";                                                    nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, HIST_ids( 9), 'long_name',   nn, long_name);                       call cke(sub,ret,1111);
     ret=nf_put_att_real(ncid, HIST_ids( 9), '_FillValue',NF_REAL, 1, FILLNC);                     call cke(sub,ret,1113);
     ret=nf_put_att_text(ncid, HIST_ids( 9), 'units', 7, 'decibar');                               call cke(sub,ret,1123);
    ! 10i-----------------------------------------------------------------
     ret=nf_def_var(ncid, 'HISTORY_STOP_PRES',      NF_REAL, NDIMS2, HIST_dim2, HIST_ids(10));    call cke(sub,ret,1120);
    long_name = "Stop pressure action applied on";                                                     nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, HIST_ids(10), 'long_name',   nn, long_name);                       call cke(sub,ret,1121);
     ret=nf_put_att_real(ncid, HIST_ids(10), '_FillValue',NF_REAL, 1, FILLNC);                   call cke(sub,ret,1123);
     ret=nf_put_att_text(ncid, HIST_ids(10), 'units', 7, 'decibar');                               call cke(sub,ret,1123);
    ! 11r-----------------------------------------------------------------
     ret=nf_def_var(ncid, 'HISTORY_PREVIOUS_VALUE',  NF_REAL,NDIMS2, HIST_dim2, HIST_ids(11));    call cke(sub,ret,1090);
    long_name = "Parameter/Flag previous value before action";                                      nn = len(trim(long_name))
     ret=nf_put_att_text(ncid, HIST_ids(11), 'long_name',   nn, long_name);                       call cke(sub,ret,1091);
     ret=nf_put_att_real(ncid, HIST_ids(11), '_FillValue',   NF_REAL, 1, FILLNC);                        call cke(sub,ret,1093);
    ! 12c-----------------------------------------------------------------
    HIST_dim3(1) = STRING16_dimid; ! char HISTORY_QCTEST           (N_HISTORY, STRING16);
     ret=nf_def_var(ncid, 'HISTORY_QCTEST',          NF_CHAR,NDIMS3, HIST_dim3, HIST_ids(12));    call cke(sub,ret,1130);
    long_name = "Documentation of tests performed, tests failed (in hex form)";                     nn = len(trim(long_name))
    conventns = "Write tests performed when ACTION=QCP$; tests failed when ACTION=QCF$";            nc = len(trim(conventns))
     ret=nf_put_att_text(ncid, HIST_ids(12), 'long_name',   nn, long_name);                       call cke(sub,ret,1131);
     ret=nf_put_att_text(ncid, HIST_ids(12), 'conventions', nc, conventns);                       call cke(sub,ret,1132);
     ret=nf_put_att_text(ncid, HIST_ids(12), '_FillValue',   1, ' ');                             call cke(sub,ret,1133);
    !--------------------------------------------------------------------

    if(DBLEV==37)then; write(msg,'("LEV37 Leave sub ",a)')sub; call wrtdbg(sub,msg); endif

end subroutine def_ncbiop_calhis
  !====================================================================
  ! SCIENTIFIC_CALIB puts
  !====================================================================
subroutine put_ncbiop_calhis(caller,ncid)

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    integer :: ihst, ipar, ical ! Counters for the dimensions
    character(len=14) :: c14;
    character(len=16) :: c16;
    character(len=64) :: c64;
    character(len=256) :: c256;		! SCICAL_<256 blank characters>
    character(len=256) :: b256;
    character(len=256) :: none;		! SCICAL_eqnt for unadjusted parameters


    character(len=16) :: perf_tmp, fail_tmp

    integer :: st2(2), ct2(2), st3(3), ct3(3), st4(4), ct4(4); ! start/count for vara uts
    ! integer :: st4(4), ct4(4); ! start/count for vara puts
    integer :: nhist
    integer :: ipf

    character(len=16) :: qcperf(MAXPRF), qcfail(MAXPRF);

    character(len=18), parameter :: sub = 'put_ncbiop_calhis'

    if(DBLEV==37)then; write(msg,'("LEV37 Enter sub ",a)')sub; call wrtdbg(sub,msg); endif

    c14 = "              ";
    c16 = "                ";
    write(c64,'(a64)')' ';
    write(c256,'(a256)')' ';
    write(b256,'(a256)')' ';
    write(none,'("none",a252)')' ';

    !==================================================================
    N_PROF = nprf;
    !==================================================================
    do iprf=1,nbprf;
	do ical = 1,N_CALIB
	    do ipar = 1,nbparout ! NOT N_PARAM - that is for the R-files
		st4(1) =    1; ct4(1) = 64; ! string length
		st4(2) = ipar; ct4(2) =  1; 
		st4(3) =    1; ct4(3) =  1; ! So far, we only have N_CALIB=1;
		st4(4) = iprf; ct4(4) =  1; 
		BPARAMETER = BPARAMETERS(ipar,iprf);

		 ret=nf_put_vara_text(ncid, SCICAL_ids(1),st4,ct4,BPARAMETER);         call cke(sub,ret,1);

		if(DBLEV==37)then; write(msg,'("PUT_SCICAL PAR PARAMETER: >",a,"<")')PARAMETER; call wrtdbg(sub,msg); endif

		SCICAL_eqtn = none; SCICAL_coef = none; SCICAL_cmnt = b256; SCICAL_date = c14;
		if(ncparout(ipar)=="PRES")then
		    SCICAL_cmnt = "Adjusted values are provided in the core profile file";
		else
		    SCICAL_cmnt = b256;
		endif
		ct4(1) = 256;  ret=nf_put_vara_text(ncid, SCICAL_ids(2),st4,ct4,SCICAL_eqtn); call cke(sub,ret,2);
		ct4(1) = 256;  ret=nf_put_vara_text(ncid, SCICAL_ids(3),st4,ct4,SCICAL_coef); call cke(sub,ret,3);
		ct4(1) = 256;  ret=nf_put_vara_text(ncid, SCICAL_ids(4),st4,ct4,SCICAL_cmnt); call cke(sub,ret,4);
		ct4(1) = 14;   ret=nf_put_vara_text(ncid, SCICAL_ids(5),st4,ct4,SCICAL_date); call cke(sub,ret,5);
	    enddo
	enddo
    enddo
    !==================================================================
    ! HISTORY puts
    !================================================================================
    write(msg,'("!==================================================================")'); call wrtdbg(sub,msg);
    c16 = "                "; ! Clear after SCICAL puts
    ! These are the same for all history records
    NHIST_inst = "AO  ";		! HISTORY_INSTITUTION
    NHIST_step = "ARGQ";		! HISTORY_STEP
    NHIST_soft = "QCPL";                ! HISTORY_SOFTWARE
    NHIST_sfrl = "    ";		! HISTORY_SOFTWARE_RELEASE
    NHIST_refr = c64;			! HISTORY_REFERENCE
    NHIST_date = file_date_time;	! HISTORY_DATE
    NHIST_actn = "QCP$";                ! HISTORY_ACTION
    NHIST_parm = "                ";	! HISTORY_PARAMETER
    NHIST_strt = FILLNC;		! HISTORY_START_PRES
    NHIST_stop = FILLNC;		! HISTORY_STOP_PRES
    NHIST_prev = FILLNC;		! HISTORY_PREVIOUS_VALUE

NHIST_actn = "QCP$"; NHIST_qctest = qcperf(iprf);

    ! Enter the DOXY history
    nhist = 1;
    NHIST_parm = "DOXY";    ! HISTORY_PARAMETER
    NHIST_actn = "QCP$";;
    do iprf=1,nbprf;
	NHIST_qctest = qcperf(iprf);
        call qctest2hex(sub,iprf,perf_tmp,fail_tmp);
        qcperf(iprf) = perf_tmp;
        qcfail(iprf) = fail_tmp;
	if(DEBUG) write(DBG,'("DBT HIST PUT @1 iprf: ",i2," nhist: ",i2)')iprf,nhist
	call put_ncbiop_histrec(sub,ncid,nhist,iprf)
    enddo
    nhist = 2;
    NHIST_actn = "QCF$";
    do iprf=1,nbprf;
	NHIST_qctest = qcfail(iprf);
	if(DEBUG) write(DBG,'("DBT HIST PUT @2 iprf: ",i2," nhist: ",i2)')iprf,nhist
	call put_ncbiop_histrec(sub,ncid,nhist,iprf)
    enddo

    if(DBLEV==37)then; write(msg,'("LEV37 Leave sub ",a)')sub; call wrtdbg(sub,msg); endif

end subroutine put_ncbiop_calhis
!================================================================================
subroutine put_ncbiop_histrec(caller,ncid,ihist,iprof)
! Jan 2018 - currently, we only need history recs for DOXY itself.  all others not needed
! At somepoint, however, we will need to QC the TEMP_DOXY field, but not quite yet

    character(*), intent( in ) :: caller        ! who is calling me?
    integer,      intent( in ) :: ncid          ! file id
    integer,      intent( in ) :: ihist		! history record to write
    integer,      intent( in ) :: iprof		! profile to write

    integer :: iret
    integer :: ret = 0;
    integer ::  loc
    integer :: st2(2), ct2(2), st3(3), ct3(3), st4(4), ct4(4); ! start/count for vara uts
    character(len=18), parameter :: sub = 'put_ncbiop_histrec'

    ! st2/ct2 are start/count for the three REAL fields to be put
    st2(1) = iprof; ct2(1) = 1;		
    st2(2) = ihist; ct2(2) = 1;		
    ! st3/ct3 are start/count for the char fields - the count of the first dimensin is the size of the string
    st3(1) = 1; !varies by string length
    st3(2) = iprof; ct3(2) = 1;		
    st3(3) = ihist; ct3(3) = 1;		
    ! NOTE NF_UNLIMITED This is the call that increments the UNLIMITED dimension for prof
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 1),st3,ct3,NHIST_inst);   call cke(sub,ret,2101);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 2),st3,ct3,NHIST_step);   call cke(sub,ret,2102);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 3),st3,ct3,NHIST_soft);   call cke(sub,ret,2103);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 4),st3,ct3,NHIST_sfrl);   call cke(sub,ret,2104);
    ct3(1) = 64;  ret=nf_put_vara_text(ncid,HIST_ids( 5),st3,ct3,NHIST_refr);   call cke(sub,ret,2105);
    ct3(1) = 14;  ret=nf_put_vara_text(ncid,HIST_ids( 6),st3,ct3,NHIST_date);   call cke(sub,ret,2106);
    ct3(1) =  4;  ret=nf_put_vara_text(ncid,HIST_ids( 7),st3,ct3,NHIST_actn);   call cke(sub,ret,2107);
    ct3(1) = 16;  ret=nf_put_vara_text(ncid,HIST_ids( 8),st3,ct3,NHIST_parm);   call cke(sub,ret,2108);
                  ret=nf_put_vara_real(ncid,HIST_ids( 9),st2,ct2,NHIST_strt);   call cke(sub,ret,2109);
                  ret=nf_put_vara_real(ncid,HIST_ids(10),st2,ct2,NHIST_stop);   call cke(sub,ret,2110);
                  ret=nf_put_vara_real(ncid,HIST_ids(11),st2,ct2,NHIST_prev);   call cke(sub,ret,2111);
    ct3(1) = 16;  ret=nf_put_vara_text(ncid,HIST_ids(12),st3,ct3,NHIST_qctest); call cke(sub,ret,2112);

    if(DBLEV==37)then; write(msg,'("LEV37 Leave sub ",a)')sub; call wrtdbg(sub,msg); endif

end subroutine put_ncbiop_histrec

!================================================================================

  !====================================================================
subroutine put_ncbiop_param(caller,ncid,ipar,iprf,PAR,BPAR_ids);


    character(*), intent( in )     :: caller
    integer, intent( in )          :: ncid       ! LNetCDF file id
    integer, intent( in )          :: ipar       ! PARAMETER number
    integer, intent( in )          :: iprf       ! PROFILE number
    character(*), intent( in )      :: PAR       ! PARAMETER name to load (only used for annnotation)
    integer, intent( in )          :: BPAR_ids(5) ! PARAMETER nc ids;

    integer :: iret
    integer :: ret = 0;
    integer :: loc
    integer :: ireg(MAXPRF), iadj(MAXPRF);
    integer :: count(2), start(2)
    real    :: DATAREG(MAXLEV); character(len=1) :: FLAGREG(MAXLEV);
    real    :: DATAADJ(MAXLEV); character(len=1) :: FLAGADJ(MAXLEV);

    character(len=16), parameter   :: sub = 'put_ncbiop_param';

    count(1) = nlev; count(2) = 1;
    start(1) = 1;    start(2) = 1;	! Note that start(2) is the profile number (iprf)

    ! Note that for biop, ONLY DOXY needs QC in adition to data - also, noladjustments, so fill
    ! data & qc for DOXY, and only data for all others

    start(2) = iprf; 
     ret=nf_put_vara_real(ncid, BPAR_ids(1), start, count, BDATA); call cke(sub,ret,301);
    if(ipar == 1) return; ! Only put thre data itself for PRES, ipar=1
    if(ipar == 2)then
	 ret=nf_put_vara_text(ncid, BPAR_ids(2), start, count, BFLAG); call cke(sub,ret,302); ! DOXY uses REAL qc flags
   else
	do ilev=1,nlev; if(BDATA(ilev) == FILLNC)then; BFLAG(ilev) = " "; else; BFLAG(ilev) = "0"; endif; enddo
	 ret=nf_put_vara_text(ncid, BPAR_ids(2), start, count, BFLAG); call cke(sub,ret,303);
    endif
    	 ret=nf_put_vara_real(ncid, BPAR_ids(3), start, count, FDATA); call cke(sub,ret,304);
	 ret=nf_put_vara_text(ncid, BPAR_ids(4), start, count, FFLAG); call cke(sub,ret,305);
    	 ret=nf_put_vara_real(ncid, BPAR_ids(5), start, count, FDATA); call cke(sub,ret,306);

end subroutine put_ncbiop_param

  !====================================================================
subroutine def_ncbiop_param_doxy(caller,ncid,PAR,PAR_ids);
character(*), intent( in )     :: caller
integer, intent( in )          :: ncid          ! LNetCDF file id
character(*), intent( in )      :: PAR          ! PARAMETER to define
integer, intent(out )          :: PAR_ids(5);   ! The PRES parameter needs only one par_id

integer :: DATA_dims(NDIMS2);
character(len=64) :: nam
                             !123456789012345678901234567890123456789012
character(len=64) :: lname = "Dissolved oxygen";                            integer :: llname = 16;
character(len=64) :: sname = "moles_of_oxygen_per_unit_mass_in_sea_water";  integer :: lsname = 42;
real              :: FillV = FILLNC;
character(len=16) :: units = "micromole/kg";                                integer :: lunits = 12;
real              :: v_min = -5.0;
real              :: v_max = 600.0;
character(len=9)  :: C_fmt = "%7.1f";                                       integer :: lC_fmt =  5;
character(len=9)  :: F_fmt = "F7.1";                                        integer :: lF_fmt =  4;
real              :: resln = 0.1;

                                !123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
character(len=86) :: lnameerr = "Contains the error on the adjusted values as determined by the delayed mode QC process";


integer :: iret
integer :: ret = 0;
integer :: loc
character(len=21), parameter   :: sub = 'def_ncbiop_param_doxy';

DATA_dims(2)         = N_PROF_dimid;  DATA_dims(1)         = N_LEVELS_dimid;

nam=trim(PAR);                     ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(1)); call cke(sub,ret,1);
 ret=nf_put_att_text(ncid, PAR_ids(1), 'long_name',     llname,     lname);
 ret=nf_put_att_text(ncid, PAR_ids(1), 'standard_name', lsname,     sname);
 ret=nf_put_att_real(ncid, PAR_ids(1), '_FillValue',    NF_REAL, 1, FillV);
 ret=nf_put_att_text(ncid, PAR_ids(1), 'units',         lunits,     units);
 ret=nf_put_att_real(ncid, PAR_ids(1), 'valid_min',     NF_REAL, 1, v_min);
 ret=nf_put_att_real(ncid, PAR_ids(1), 'valid_max',     NF_REAL, 1, v_max);
 ret=nf_put_att_text(ncid, PAR_ids(1), 'C_format',      lC_fmt,     C_fmt);
 ret=nf_put_att_text(ncid, PAR_ids(1), 'FORTRAN_format',lF_fmt,     F_fmt);
 ret=nf_put_att_real(ncid, PAR_ids(1), 'resolution',    NF_REAL, 1, resln);

nam=trim(PAR)//"_QC";              ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS2,DATA_dims,PAR_ids(2)); call cke(sub,ret,1);
 ret=nf_put_att_text(ncid, PAR_ids(2), 'long_name',   12, 'quality flag' );
 ret=nf_put_att_text(ncid, PAR_ids(2), 'conventions', 22, 'Argo reference table 2' );
 ret=nf_put_att_text(ncid, PAR_ids(2), '_FillValue',   1, ' ');

nam=trim(PAR)//"_ADJUSTED";        ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(3)); call cke(sub,ret,1);
 ret=nf_put_att_text(ncid, PAR_ids(3), 'long_name',     llname, lname);
 ret=nf_put_att_text(ncid, PAR_ids(3), 'standard_name', lsname, trim(sname));
 ret=nf_put_att_real(ncid, PAR_ids(3), '_FillValue',    NF_REAL, 1, FillV);
 ret=nf_put_att_text(ncid, PAR_ids(3), 'units',         lunits, units);
 ret=nf_put_att_real(ncid, PAR_ids(3), 'valid_min',     NF_REAL, 1,v_min);
 ret=nf_put_att_real(ncid, PAR_ids(3), 'valid_max',     NF_REAL, 1,v_max);
 ret=nf_put_att_text(ncid, PAR_ids(3), 'C_format',      lC_fmt, C_fmt);
 ret=nf_put_att_text(ncid, PAR_ids(3), 'FORTRAN_format',lF_fmt, F_fmt);
 ret=nf_put_att_real(ncid, PAR_ids(3), 'resolution',    NF_REAL, 1, resln);

nam=trim(PAR)//"_ADJUSTED_QC";     ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS2,DATA_dims,PAR_ids(4)); call cke(sub,ret,1);
 ret=nf_put_att_text(ncid, PAR_ids(4), 'long_name',   12, 'quality flag' );
 ret=nf_put_att_text(ncid, PAR_ids(4), 'conventions', 22, 'Argo reference table 2' );
 ret=nf_put_att_text(ncid, PAR_ids(4), '_FillValue',   1, ' ');

nam=trim(PAR)//"_ADJUSTED_ERROR";  ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(5)); call cke(sub,ret,1);
 ret=nf_put_att_text(ncid, PAR_ids(5), 'long_name',     len(trim(lnameerr)), lnameerr);
 ret=nf_put_att_real(ncid, PAR_ids(5), '_FillValue',    NF_REAL, 1, FillV);
 ret=nf_put_att_text(ncid, PAR_ids(5), 'units',         lunits,     units);
 ret=nf_put_att_text(ncid, PAR_ids(5), 'C_format',      lC_fmt, C_fmt);
 ret=nf_put_att_text(ncid, PAR_ids(5), 'FORTRAN_format',lF_fmt, F_fmt);
 ret=nf_put_att_real(ncid, PAR_ids(5), 'resolution',    NF_REAL, 1, resln);

end subroutine def_ncbiop_param_doxy


  !====================================================================
subroutine def_ncbiop_param(caller,ncid,ipar,PAR,PAR_ids);
character(*), intent( in )     :: caller
integer, intent( in )          :: ncid          ! LNetCDF file id
integer, intent( in )          :: ipar      ! PARAMETER number
character(*), intent( in )      :: PAR          ! PARAMETER to define
integer, intent(out )          :: PAR_ids(5);   ! The PRES parameter needs only one par_id

character(len=64) :: lname; integer :: llname; ! long_name and its length
character(len=64) :: sname; integer :: lsname; ! standard)name and its length
character(len=64) :: units; integer :: lunits; ! units and its length
! character(len=64) :: ; integer :: l; !   and its length

integer :: DATA_dims(NDIMS2);
character(len=48) :: nam

real              :: FillV = FILLNC;
real              :: v_min = 0.0;
real              :: v_max = 2000.0;
character(len=5)  :: C_fmt = "%7.1f";                                     integer :: lC_fmt =  5;
character(len=4)  :: F_fmt = "F7.1";                                      integer :: lF_fmt =  4;
real              :: resln = 0.1;

character(len=87) :: err_lname = 'Contains the error on the adjusted values as determined by the delayed mode QC process';

integer :: iret
integer :: ret = 0;
integer :: loc
logical :: usesname; ! Do we need these (standard_name)
logical :: usemnmx;  ! Do we need valid_min/max?
character(len=16), parameter   :: sub = 'def_ncbiop_param';

DATA_dims(2)         = N_PROF_dimid;  DATA_dims(1)         = N_LEVELS_dimid;

nam=trim(PAR);                     ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(1)); call cke(sub,ret,1);
nam=trim(PAR)//"_QC";              ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS2,DATA_dims,PAR_ids(2)); call cke(sub,ret,1);
nam=trim(PAR)//"_ADJUSTED";        ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(3)); call cke(sub,ret,1);
nam=trim(PAR)//"_ADJUSTED_QC";     ret=nf_def_var(ncid,nam,NF_CHAR,NDIMS2,DATA_dims,PAR_ids(4)); call cke(sub,ret,1);
nam=trim(PAR)//"_ADJUSTED_ERROR";  ret=nf_def_var(ncid,nam,NF_REAL,NDIMS2,DATA_dims,PAR_ids(5)); call cke(sub,ret,1);

usesname = BPAR_use_snam(ipar);
usemnmx = BPAR_mnmx(ipar); if(usemnmx)then; v_min = BPAR_vmin(ipar); v_max = BPAR_vmax(ipar); endif

             lname = BPAR_lname(ipar); llname = len(trim(lname));
if(usesname) sname = BPAR_sname(ipar); lsname = len(trim(sname));
             units = BPAR_units(ipar); lunits = len(trim(units));

              ret=nf_put_att_text(ncid, PAR_ids(1), 'long_name',     llname,     lname);
if(usesname)  ret=nf_put_att_text(ncid, PAR_ids(1), 'standard_name', lsname,     sname);
              ret=nf_put_att_real(ncid, PAR_ids(1), '_FillValue',    NF_REAL, 1, FillV);
              ret=nf_put_att_text(ncid, PAR_ids(1), 'units',         lunits,     units);
if(usemnmx)   ret=nf_put_att_real(ncid, PAR_ids(1), 'valid_min',     NF_REAL, 1, v_min);
if(usemnmx)   ret=nf_put_att_real(ncid, PAR_ids(1), 'valid_max',     NF_REAL, 1, v_max);
              ret=nf_put_att_text(ncid, PAR_ids(1), 'C_format',      lC_fmt,     C_fmt);
              ret=nf_put_att_text(ncid, PAR_ids(1), 'FORTRAN_format',lF_fmt,     F_fmt);
              ret=nf_put_att_real(ncid, PAR_ids(1), 'resolution',    NF_REAL, 1, resln);


              ret=nf_put_att_text(ncid, PAR_ids(2), 'long_name',   12, 'quality flag' );
              ret=nf_put_att_text(ncid, PAR_ids(2), 'conventions', 22, 'Argo reference table 2' );
              ret=nf_put_att_text(ncid, PAR_ids(2), '_FillValue',   1, ' ');

              ret=nf_put_att_text(ncid, PAR_ids(3), 'long_name',     llname, lname);
if(usesname)  ret=nf_put_att_text(ncid, PAR_ids(3), 'standard_name', lsname, trim(sname));
              ret=nf_put_att_real(ncid, PAR_ids(3), '_FillValue',    NF_REAL, 1, FillV);
              ret=nf_put_att_text(ncid, PAR_ids(3), 'units',         lunits, units);
              ret=nf_put_att_real(ncid, PAR_ids(3), 'valid_min',     NF_REAL, 1,v_min);
              ret=nf_put_att_real(ncid, PAR_ids(3), 'valid_max',     NF_REAL, 1,v_max);
              ret=nf_put_att_text(ncid, PAR_ids(3), 'C_format',      lC_fmt, C_fmt);
              ret=nf_put_att_text(ncid, PAR_ids(3), 'FORTRAN_format',lF_fmt, F_fmt);
              ret=nf_put_att_real(ncid, PAR_ids(3), 'resolution',    NF_REAL, 1, resln);

              ret=nf_put_att_text(ncid, PAR_ids(4), 'long_name',   12, 'quality flag' );
              ret=nf_put_att_text(ncid, PAR_ids(4), 'conventions', 22, 'Argo reference table 2' );
              ret=nf_put_att_text(ncid, PAR_ids(4), '_FillValue',   1, ' ');

  ret=nf_put_att_text(ncid, PAR_ids(5), 'long_name',     86, err_lname);
! None of the errors have a standard name
              ret=nf_put_att_real(ncid, PAR_ids(5), '_FillValue',    NF_REAL, 1, FillV);
              ret=nf_put_att_text(ncid, PAR_ids(5), 'units',         lunits,     units);
              ret=nf_put_att_text(ncid, PAR_ids(5), 'C_format',      lC_fmt, C_fmt);
              ret=nf_put_att_text(ncid, PAR_ids(5), 'FORTRAN_format',lF_fmt, F_fmt);
              ret=nf_put_att_real(ncid, PAR_ids(5), 'resolution',    NF_REAL, 1, resln);

end subroutine def_ncbiop_param

!================================================================================
end module ncbiop_arrays
!================================================================================  
